// Unreal Fighting Framework by Kevin Poretti

#include "FFInputBufferComponent.h"

UFFInputBufferComponent::UFFInputBufferComponent()
{
	PrimaryComponentTick.bCanEverTick = false;
}


void UFFInputBufferComponent::AddInput(const FFFInputState& InputState)
{
	InputBuffer.ForcePush(InputState);
}


bool UFFInputBufferComponent::CheckInputSequence(const FFFInputSequence& InputSequence)
{
    if(InputSequence.Sequence.IsEmpty())
    {
        UE_LOG(LogTemp, Warning, 
            TEXT("FFInputBufferComponent :: CheckInputSequence - tried to check input sequence but it was empty"));
        return false;
    }

    int CondIdx = InputSequence.Sequence.Num() - 1;
    int ElapsedFrames = 0;
    for(int InpIdx = InputBuffer.Num() - 1; InpIdx > 0; InpIdx--)
    {
        int32 RequiredButtons = InputSequence.Sequence[CondIdx].RequiredButtons;
        EFFButtonState RequiredButtonState = InputSequence.Sequence[CondIdx].RequiredButtonState;
        int32 TimeoutDuration = InputSequence.Sequence[CondIdx].TimeoutDuration;
        int32 PrevInput = InputBuffer[InpIdx - 1].Buttons;
        int32 CurrInput = InputBuffer[InpIdx].Buttons;
        int32 PrevDisable = InputBuffer[InpIdx - 1].DisabledButtons;
        int32 CurrDisable = InputBuffer[InpIdx].DisabledButtons;
        bool bCondFoundThisIteration = false;
        switch (RequiredButtonState)
        {
            // TODO: should it be (PrevInput & RequiredButtons) == RequiredButtons or what we have now?
            // TODO: If you have a sequence that is defined as something like button 1 pressed + button 2 pressed
            // and the buttons are pressed on the same exact frame, then this implementation will not
            // account for that. We should keep checking the sequences against the current frame's
            // inputs until no sequences are detected, then continue on checking through the input buffer.
            case EFFButtonState::BTNS_Pressed:
                if(!(PrevInput & RequiredButtons | PrevDisable) && 
                   CurrInput & RequiredButtons & ~CurrDisable)
                {
                    CondIdx--;
                    ElapsedFrames = 0;
                    bCondFoundThisIteration = true;
                }
                break;
            case EFFButtonState::BTNS_Released:
                if(PrevInput & RequiredButtons & ~PrevDisable &&
                   !(CurrInput & RequiredButtons | CurrDisable))
                {
                    CondIdx--;
                    ElapsedFrames = 0;
                    bCondFoundThisIteration = true;
                }
                break;
            case EFFButtonState::BTNS_Down:
                if(CurrInput & RequiredButtons & ~CurrDisable)
                {
                    CondIdx--;
                    ElapsedFrames = 0;
                    bCondFoundThisIteration = true;
                }
                break;
            // TODO: implement button held condition
            /*
            case EFFButtonState::BTNS_Held:
                break;
            */
        }

        // All conditions were met
        if(CondIdx == -1)
        {
            return true;
        }

        // No condition found this iteration, so increment elapse frames and see if current condition timed out
        if(!bCondFoundThisIteration)
        {
            ElapsedFrames++;
            if(ElapsedFrames > TimeoutDuration)
            {
                return false;
            }
        }
    }

	return false;
}

void UFFInputBufferComponent::DisableMostRecentInput()
{
    if(InputBuffer.Num() < 2)
    {
        return;
    }
    
    InputBuffer[InputBuffer.Num() - 1].DisabledButtons |= InputBuffer[InputBuffer.Num() - 1].Buttons;
    InputBuffer[InputBuffer.Num() - 2].DisabledButtons |= InputBuffer[InputBuffer.Num() - 2].Buttons;
}

void UFFInputBufferComponent::FlushBuffer()
{
    InputBuffer.Flush();
}
