// Unreal Fighting Framework by Kevin Poretti

#pragma once

// FF includes
#include "Input/FFInputBufferComponent.h"

// UE includes
#include "CoreMinimal.h"

#include "FFState.generated.h"

USTRUCT(BlueprintType)
struct FFFStateContext
{
    GENERATED_BODY()

	/** Actor that owns the avatar. Typically a player controller. */
	UPROPERTY(BlueprintReadOnly)
	AActor* Owner;

	/**
	 *	Actor that represents the player's avatar or an object associated with the player's avatar.
	 *	This is typically a character or a weapon.
	 */
	UPROPERTY(BlueprintReadOnly)
	AActor* Avatar;

    /**
     * Parent state machine that controls this state
     */
	UPROPERTY(BlueprintReadOnly)
    class UFFStateMachineComponent* Parent;
};

DECLARE_DYNAMIC_DELEGATE_OneParam(FFFInputEventDelegate, const FFFStateContext&, InStateContext);

/*
 * TODO: check if this is even used anymore and if not delete? I recall using this in a few
 * states/actions but it was a pain to use and I switched to checking an input sequence directly in
 * a state's update function
 */
USTRUCT(BlueprintType)
struct FFFInputEventHandler
{
    GENERATED_BODY()

	// TODO: document
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly)
	FFFInputSequence RequiredSequence;

	// TODO: document
    UPROPERTY(EditDefaultsOnly, BlueprintReadOnly)
	FFFInputEventDelegate Delegate;
};

UENUM(BlueprintType)
enum class EFFStateFinishReason : uint8
{
	/** State ended due to another state's entry condition being satisfied  */
	SFT_Interrupted				UMETA(DisplayName="Interrupted"),

	/** State ended because the number of ticks in this state exceeded state's duration property */
    SFT_DurationMetOrExceeded			UMETA(DisplayName="Duration Reached"),
	
	/** State ended because the owner is no longer in the movement mode required by the state */
    SFT_NotInReqMovementMode	UMETA(DisplayName="Not In Required Movement Mode"),
    
	SFT_MAX						UMETA(Hidden)
};


/**
 *	A state is an object that provides rules and conditions for when a state can be transitioned into
 *	and logic to run when the state is entered, exited, and active.
 */
UCLASS(Blueprintable)
class UNREALFIGHTINGFRAMEWORK_API UFFState : public UObject
{
	GENERATED_BODY()

public:
	// State parameters (should be read-only)
    /** Name of this state behavior */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="UFF State Properties")
	FName Name;

    /**
	 *	True if this state has some duration where it should return to the entry state
	 *	if this state is not cancelled out of by other means.
	 */
    UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="UFF State Properties")
	bool bStateHasDuration;

	// TODO: this should only be editable if bStateHasDuration is true
	/**
	 *	How long this state will be active before finishing if this state is not cancelled out of
	 *	by other means.
	 */
    UPROPERTY(EditDefaultsOnly, BlueprintReadWrite, Category="UFF State Properties")
	int64 StateDuration;

	/**
	 *	What stance the object this state represents must be in.
	 *	For example this is usually an enumerated value that signifies if a character needs to be
	 *	crouching, standing, or airborne for this state to be eligible for transitioning.
	 */
    UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="UFF State Properties")
	int32 StanceRequired;

	/**
	 *	What is this state's category.
	 *	Used for determining what types of state can prematurely cancel this one.
	 */
    UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="UFF State Properties")
	int32 StateType;

	/**
	 *	Input sequences that needs to be present in the controlling player's input buffer for this
	 *	state to be eligible for transitioning.
	 */
    UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="UFF State Properties")
	TArray<FFFInputSequence> InputSequences;

	/**
	 *	Should the most recent input in the controlling player's input buffer be disabled when
	 *	entering this state?
	 */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="UFF State Properties")
	bool bDisableMostRecentInputOnEntry = true;
	
	/**
	 *	If true the state can transition from itself into itself without having to go through
	 *	another state like Idle
	 */
    UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="UFF State Properties")
	bool bCanTransitionToSelf = false;

	/**
	 *	Specifies whether a state is a followup state or not.
	 *
	 *	Followup states are states that should only be transitioned into from another specific state
	 *	but otherwise have no other specific conditions that need to be met to transition.
	 *	Setting this flag to true when there are no other conditions prevents this state from
	 *	constantly being transitioned into by the state machine.
	 */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="UFF State Properties")
	bool bIsFollowupState = false;
	
	/**
	 *	State to transition to when this state is finished. If left blank then the entry state
	 *	of the state machine will be transitioned into.
	 */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="UFF State Properties")
	FName FollowupState = NAME_None;
	
	/**
	 *	Animation to begin playing when this state is entered
	 */
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="UFF State Properties")
	UAnimMontage* MontageToPlay;

	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="UFF State Properties")
	bool bStopMontageOnStateEnd;

    UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="UFF State Properties")
	bool bStopMontageOnMovementModeChange;

	/**
	 *	Event delegates to call when a certain input condition is detected in the Owner's input buffer
	 */
	UPROPERTY(BlueprintReadOnly, Category="UFF State Events")
	TArray<FFFInputEventHandler> InputHandlers;

    /**
	 *	Called when state is first created.
	 */
	virtual void Init(const FFFStateContext& InStateContext);

	/**
	 *	Returns true if Avatar's is in the correct stance AND
	 *	the state type is enabled (or the state can be hit or whiff cancelled from the current state) AND
	 *	all state entry conditions were met AND
	 *	at least one input sequence is present in the Owner's input buffer
	 */ 
	virtual bool CanTransition(const FFFStateContext& InStateContext);

    /**
	 *	Called whenever this state is transitioned into.
	 *
	 *	Calls appropriate Blueprint hooks.
	 */
	virtual void Enter(const FFFStateContext& InStateContext);

    /**
	 *	Called whenever this state is transitioned out of into a new state.
	 *
	 *	Calls appropriate Blueprint hooks.
	 *
	 *	@param StateFinishReason the reason the state is ending i.e. the state's duration elapse
	 */
	virtual void Exit(const FFFStateContext& InStateContext, EFFStateFinishReason StateFinishReason);

    /**
	 *	Called whenever this state is active and the game logic ticks.
	 *
	 *	Calls appropriate Blueprint hooks.
	 *
	 *	@param OneFrame the time that elapses during one fixed tick
	 */
    virtual void Update(float OneFrame, const FFFStateContext& InStateContext);

	/**
	 *	Called whenever state owners lands on a walkable surface after they were previously falling.
	 *
	 *	Calls appropriate Blueprint hooks.
	 *
	 *	@param Hit Hit result for the surface that was landed on
	 */
	virtual void Landed(const FHitResult& Hit, const FFFStateContext& InStateContext);

	/**
	 *	@brief Called whenever the state's owner current movement mode changes
	 *
	 *	Calls appropriate Blueprint hooks.
	 *
	 *	@param PrevMovementMode previous movement mode.
	 *	@param PreviousCustomMode previous movement mode if PrevMovementMode is MOVE_Custom
	 *	@param NewMovementMode new movement mode
	 *	@param NewCustomMode new movement mode if NewMovementMode is MOVE_Custom
	 */ 
	virtual void MovementModeChanged(EMovementMode PrevMovementMode, uint8 PreviousCustomMode, 
	    EMovementMode NewMovementMode, uint8 NewCustomMode, const FFFStateContext& InStateContext);

	// TODO: document
	virtual void AttackHit(const FHitResult& HitResult, const FFFStateContext& InStateContext);
	
	// TODO: document
	// TODO: call this callback when the avatar is hit
	// TODO: pass in hitdata struct as well
    virtual void HitTaken(/*const FHitParams& HitParams,*/ const FFFStateContext& InStateContext);

	// TODO: document
	// TODO: call this callback when the avatar blocks a hit
	// TODO: pass in hitdata struct as well
	virtual void BlockTaken(const FFFStateContext& InStateContext);

    /**
	 *	Called when you want to exit from this state but no eligible transitions exist to other states,
	 *	usually due to a lack of player input.
	 *
	 *	By default this will transition to the owning state machine's entry state, which is usually
	 *	an idle state.
	 *
	 *	This function will get called automatically when the ticks in this state reaches the
	 *	threshold set by the StateDuration member, but you can call it whenever you wish to return
	 *	to the entry state.
	 *
     *	@param InStateContext 
     */
    UFUNCTION(BlueprintCallable)
	virtual void Finish(const FFFStateContext& InStateContext, EFFStateFinishReason StateFinishReason);
	
	/** @brief Blueprint hook for when the state is first created. */
	UFUNCTION(BlueprintNativeEvent, Category="UFF|State|Events")
	void OnInit(const FFFStateContext& InStateContext);

	/**
	 *	Blueprint hook for overriding the CanTransition logic
	 */
    UFUNCTION(BlueprintNativeEvent, Category="UFF|State|Events")
	bool OnCanTransition(const FFFStateContext& InStateContext);

    /**
     *	Blueprint hook for whenever this state is transitioned into
     */
	UFUNCTION(BlueprintImplementableEvent, Category="UFF|State|Events")
	void OnEnter(const FFFStateContext& InStateContext);

	/**
	 *	Blueprint hook for whenever this state is transitioned out of into a new state
	 */
	UFUNCTION(BlueprintImplementableEvent, Category="UFF|State|Events")
	void OnExit(const FFFStateContext& InStateContext);

	/**
	 *	Blueprint hook for whenever this state is active and the game logic ticks.
	 *
	 *	@param OneFrame the time that elapses during one fixed tick
	 */
	UFUNCTION(BlueprintImplementableEvent, Category="UFF|State|Events")
	void OnUpdate(float OneFrame, const FFFStateContext& InStateContext);

	/**
	 *	Returns the next state to transition into when this state finishes.
	 *
	 *	This is called during the Finish function, either when the state duration is reached or the
	 *	state is manually finished due to some other logic.
	 *
	 *	By default this returns the name of the state specified by the FollowupState property or
	 *	NAME_None if the FollowupState is not specified.
	 */
	UFUNCTION(BlueprintNativeEvent, Category="UFF|State|Events")
	FName GetFollowupState(const FFFStateContext& InStateContext);
	
	// TODO: Don't like this at all and needs to be refactored or redesigned
	UFUNCTION(BlueprintCallable, Category="UFF|State|Animations")
	void PlayMontage(const FFFStateContext& InStateContext);

	// TODO: Don't like this at all and needs to be refactored or redesigned
    /**
     *	Blueprint hook for overriding the logic for when a anim montage plays at the start of a state
     *	@param InStateContext 
     */
    UFUNCTION(BlueprintNativeEvent, Category="UFF|State|Events")
	void OnPlayMontage(const FFFStateContext& InStateContext);
	
	/**
	 *	@brief Blueprint hook for whenever state owners lands on a walkable surface after they were
	 *	previously falling.
	 *	
	 *	@param Hit Hit result for the surface that was landed on
	 */
	UFUNCTION(BlueprintImplementableEvent, Category="UFF|State|Events")
	void OnLanded(const FHitResult& Hit, const FFFStateContext& InStateContext);

	/**
	 *	@brief Blueprint hook whenever the state's owner current movement mode changes
	 *	
	 *	@param PrevMovementMode previous movement mode.
	 *	@param PreviousCustomMode previous movement mode if PrevMovementMode is MOVE_Custom
	 *	@param NewMovementMode new movement mode
	 *	@param NewCustomMode new movement mode if NewMovementMode is MOVE_Custom
	 */
	UFUNCTION(BlueprintImplementableEvent, Category="UFF|State|Events")
	void OnMovementModeChanged(EMovementMode PrevMovementMode, uint8 PreviousCustomMode, 
	    EMovementMode NewMovementMode, uint8 NewCustomMode, const FFFStateContext& InStateContext);

	// TODO: document
    // TODO: pass in hitdata struct as well
    UFUNCTION(BlueprintImplementableEvent, Category="UFF|State|Events")
	void OnAttackHit(const FHitResult& Hit, const FFFStateContext& InStateContext);

	// TODO: document
	UFUNCTION(BlueprintImplementableEvent, Category="UFF|State|Events")
	void OnHitTaken(/*const FHitParams,*/ const FFFStateContext& InStateContext);
	
	// TODO: document
    // TODO: pass in hitdata struct as well
    UFUNCTION(BlueprintImplementableEvent, Category="UFF|State|Events")
	void OnBlockTaken(/*const FHitParams,*/ const FFFStateContext& InStateContext);

	// UObject interface
	virtual UWorld* GetWorld() const override;
	// End of UObject interface

private:
	void StopCurrentAnimMontage(const FFFStateContext& InStateContext);
};
