// Unreal Fighting Framework by Kevin Poretti

#pragma once

// FF includes
#include "IFFSystemInterface.h"
#include "FFState.h"

// UE includes
#include "CoreMinimal.h"
#include "Components/ActorComponent.h"

#include "FFStateMachineComponent.generated.h"

DECLARE_DYNAMIC_MULTICAST_DELEGATE_TwoParams(FOnStateTransitionSignature, 
    const UFFState*, PrevState, const UFFState*, NextState);

/**
 *	A state machine is a component that evaluates and controls the transitions for state objects that
 *	are a part of this state machine.
 *
 *	This component also calls the appropriate state logic when a state is changed or the component ticks.
 */
UCLASS( ClassGroup=(UnrealFightingFramework), meta=(BlueprintSpawnableComponent) )
class UNREALFIGHTINGFRAMEWORK_API UFFStateMachineComponent : public UActorComponent, public IFFSystemInterface
{
	GENERATED_BODY()

public:	
	UFFStateMachineComponent();

	/**
	 *	Creates and adds default states and enters the first state
	 */
	void Initialize(TSubclassOf<UFFState> EntryState, const TArray<TSubclassOf<UFFState>>& InitialStates);

    /**
	 *	Initializes pointers to what owns this state machine and what avatar this state represents.
	 *
	 *	These pointers will also be passed to the states owned by this state machine.
	 *
	 *	@param InOwner Actor that owns this state machine.
	 *	@param InAvatar Actor that this state machine represents.
	 */
	virtual void InitActorInfo(AActor* InOwner, AActor* InAvatar);

    /**
	 *	Creates an instance of the state class and adds newly created state to this state machine.
	 *
	 *	@param StateClassToAdd State class type to be added to this state machine
	 *
	 *	@return A pointer to the state that was added or nullptr if there was an issue adding or creating the state
	 */
	// TODO: return should probably be a const pointer if this function is public. I don't want
	// anything outside of the state machine modifying the state. Really, I want states in general to
	// be read only objects that are only modified at edit time when creating state blueprints.
	UFFState* AddState(TSubclassOf<UFFState> StateClassToAdd);

    /**
	 *	Creates an instance of the state classes and adds newly created states to this state machine.
	 *
	 *	@param StateClassesToAdd Array of state class types to be added to this state machine
	 */
	void AddStates(const TArray<TSubclassOf<UFFState>>& StateClassesToAdd);

    /**
	 *	Destroys the state with corresponding name and removes it from this state machine.
	 */
	void RemoveState(FName StateToRemove);

    /**
	 *	Transitions from CurrentState to the new state with the name passed to this function
	 *
	 *	Triggers the Exit callback on the CurrentState and the Enter callback on the new state
	 *
	 *	@param StateFinishReason the reason the current state is ending i.e. the current state's
	 *	duration elapsed
	 */
	UFUNCTION(BlueprintCallable)
	void GoToState(FName NewStateName, EFFStateFinishReason StateFinishReason);

	/**
	 *	Transitions from CurrentState to the new state passed to this function
	 *
	 *	Triggers the Exit callback on the CurrentState and the Enter callback on the new state
	 *
	 *	@param StateFinishReason the reason the current state is ending i.e. the current state's
	 *	duration elapsed
	 */
	void GoToState(UFFState* NewState, EFFStateFinishReason StateFinishReason);

	/**
	 *	Transitions from CurrentState to the default entry state
	 *
	 *	@param StateFinishReason the reason the current state is ending i.e. the current state's
	 *	duration elapsed
	 */
	void GoToEntryState(EFFStateFinishReason StateFinishReason);

    UFUNCTION(BlueprintPure)
	FORCEINLINE int64 GetTicksInState() const { return TicksInState; }

	/**
	 *	Returns the tick number which corresponds to when this state was entered into.
	 *
	 *	The tick number represents the ticks that have elapsed since the game began.
	 */
	UFUNCTION(BlueprintPure)
	FORCEINLINE int64 GetTickStateWasEntered() const { return TickStateWasEntered; }

    /**
	 *	Returns the tick number which corresponds to when the On Landed event fired
	 *
	 *	The tick number represents the ticks that have elapsed since the game began.
	 */
	UFUNCTION(BlueprintPure)
	FORCEINLINE int64 GetOnLandedTick() const { return OnLandedTick; }

	UFUNCTION(BlueprintPure)
	const UFFState* GetCurrentState() const { return const_cast<UFFState*>(CurrentState); }

    /**
     *	Returns the name of the current state
     */
	UFUNCTION(BlueprintPure)
	FORCEINLINE FName GetCurrentStateName() const;

    /**
     *	Returns the name of the current sub state label
     */
	UFUNCTION(BlueprintPure)
	FORCEINLINE FName GetCurrentSubStateLabel() const { return CurrentSubStateLabel; }

	UFUNCTION(BlueprintCallable)
	void SetSubStateLabel(FName InSubStateLabel);

    /**
	 * TODO: document
	 */
	FFFStateContext GetCurrentStateContext();

	// Events
	/**
	 *	Called whenever state machine's owner lands on a walkable surface after they were previously falling.
	 *
	 *	@param Hit Hit result for the surface that was landed on
	 */
	virtual void Landed(const FHitResult& Hit);

	/**
	 *	@brief Called whenever the state machine's owner current movement mode changes
	 *
	 *	Calls appropriate Blueprint hooks.
	 *
	 *	@param PrevMovementMode previous movement mode.
	 *	@param PreviousCustomMode previous movement mode if PrevMovementMode is MOVE_Custom
	 *	@param NewMovementMode new movement mode
	 *	@param NewCustomMode new movement mode if NewMovementMode is MOVE_Custom
	 */ 
	virtual void MovementModeChanged(EMovementMode PrevMovementMode, uint8 PreviousCustomMode, 
	    EMovementMode NewMovementMode, uint8 NewCustomMode);

	// TODO: document
	virtual void AttackHit(const FHitResult& HitResult);
	
	// TODO: document
	// TODO: bring HitParams struct into FightingFramework
	virtual void TakeHit(/* const FHitParams& HitParams*/);
	
	FOnStateTransitionSignature OnStateTransition;

	// IFFSystemInterface interface
    virtual void FixedTick(float OneFrame) override;
	// End of IFFSystemInterface interface

protected:
	/**
	 *	Actor that owns this state machine.
	 *	This will typically be a player controller that possesses the avatar.
	 */
	UPROPERTY()
	AActor* Owner;

	/**
	 *	The avatar is an actor that this state machine represents.
	 *	This will typically be a pawn or character the state machine is attached to.
	 */
	UPROPERTY()
	AActor* Avatar;

    /** How many ticks have elapsed since the currently active state was entered */
	int64 TicksInState;

    /**
	 *	The tick number which corresponds to when the current state was entered.
	 *
	 *	The tick number represents the ticks that have elapsed since the game began.
	 */
	int64 TickStateWasEntered;

    /**
	 *	The tick number which corresponds to when the On Landed event was called for the current tick.
	 *
	 *	The tick number represents the ticks that have elapsed since the game began.
	 */
	int64 OnLandedTick;

	/** Current active state for this state machine */
	UPROPERTY()
	UFFState* CurrentState;

	/** Current SubState label or NAME_None if there is no SubState label set for the current state */
	FName CurrentSubStateLabel;

	// TODO: should be a TMap
	// States that have been added
	UPROPERTY()
	TArray<UFFState*> States;

    /**
	 *	Returns the state with corresponding name 
	 */
	UFFState* FindStateWithName(FName StateName);

#if !UE_BUILD_SHIPPING
	// UActorComponent interface
	virtual void TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction) override;
	// End of UActorComponent interface
#endif
};
