// Unreal Fighting Framework by Kevin Poretti

// FF includes
#include "Input/FFInputBufferComponent.h"

// UE includes
#include "CoreMinimal.h"
#include "Misc/AutomationTest.h"

#if WITH_DEV_AUTOMATION_TESTS

IMPLEMENT_SIMPLE_AUTOMATION_TEST(FInputBufferTest, "FightingFramework.Input.InputBuffer",
	EAutomationTestFlags::ApplicationContextMask | EAutomationTestFlags::ProductFilter)
bool FInputBufferTest::RunTest(const FString& Parameters)
{
	TStrongObjectPtr<UWorld> World = TStrongObjectPtr(UWorld::CreateWorld(EWorldType::Game, false));
	// Create a dummy actor
	TStrongObjectPtr<AActor> InputBufferOwner = TStrongObjectPtr<AActor>(World->SpawnActor<AActor>());
	TestTrue(TEXT("Create InputBufferOwner"), InputBufferOwner.IsValid());
	UFFInputBufferComponent* InputBuffer =
		NewObject<UFFInputBufferComponent>(InputBufferOwner.Get(), 
		UFFInputBufferComponent::StaticClass(), TEXT("InputBuffer"));
	InputBuffer->RegisterComponent();
	
	// InputBuffer must be able to be created as a subobject of an actor
	{
		TestNotNull(TEXT("Create new InputBuffer component"), InputBuffer);
	}

	// InputBuffer should not attempt to evaluate an input sequence if there is only one input in
	// the buffer and should return false
	{
		FFFInputState InputDown;
		InputDown.Buttons = 0x00000001;
		
		InputBuffer->AddInput(InputDown);
		
		FFFInputCondition PressedCondition;
		PressedCondition.RequiredButtons = 0x00000001;
		PressedCondition.RequiredButtonState = EFFButtonState::BTNS_Pressed;
		PressedCondition.TimeoutDuration = 0;

		FFFInputSequence PressedSequence;
		PressedSequence.Sequence.Add(PressedCondition);
		TestFalse(TEXT("Don't try to detect a sequence if only one input is present in the buffer"),
			InputBuffer->CheckInputSequence(PressedSequence));
	}

	InputBuffer->FlushBuffer();

	// InputBuffer should not attempt to evaluate an input sequence if the sequence has no input conditions
	{
		FFFInputState InputUp;
		InputUp.Buttons = 0x00000000;
		FFFInputState InputDown;
		InputDown.Buttons = 0x00000001;

		InputBuffer->AddInput(InputUp);
		InputBuffer->AddInput(InputDown);

		FFFInputSequence EmptySequence;
		TestFalse(TEXT("Don't try to detect a sequence if the sequence has no input conditions"),
			InputBuffer->CheckInputSequence(EmptySequence));
	}

	InputBuffer->FlushBuffer();
	
	// InputBuffer must detect simple button pressed condition
	{
		FFFInputState InputUp;
		InputUp.Buttons = 0x00000000;
		FFFInputState InputDown;
		InputDown.Buttons = 0x00000001;

		InputBuffer->AddInput(InputUp);
		InputBuffer->AddInput(InputDown);
		
		FFFInputCondition PressedCondition;
		PressedCondition.RequiredButtons = 0x00000001;
		PressedCondition.RequiredButtonState = EFFButtonState::BTNS_Pressed;
		PressedCondition.TimeoutDuration = 0;

		FFFInputSequence PressedSequence;
		PressedSequence.Sequence.Add(PressedCondition);
		TestTrue(TEXT("Detect a single button press in the buffer"),
			InputBuffer->CheckInputSequence(PressedSequence));
	}

	InputBuffer->FlushBuffer();
	
	// InputBuffer must detect simple button released condition
	{
		FFFInputState InputUp;
		InputUp.Buttons = 0x0000000;
		FFFInputState InputDown;
		InputDown.Buttons = 0x00000001;

		InputBuffer->AddInput(InputDown);
		InputBuffer->AddInput(InputUp);
		
		FFFInputCondition ReleasedCondition;
		ReleasedCondition.RequiredButtons = 0x00000001;
		ReleasedCondition.RequiredButtonState = EFFButtonState::BTNS_Released;
		ReleasedCondition.TimeoutDuration = 0;

		FFFInputSequence ReleasedSequence;
		ReleasedSequence.Sequence.Add(ReleasedCondition);
		TestTrue(TEXT("Detect a single button release in the buffer"),
			InputBuffer->CheckInputSequence(ReleasedSequence));
	}

	InputBuffer->FlushBuffer();
	
	// InputBuffer must detect simple button down condition
	{
		FFFInputState InputUp;
		InputUp.Buttons = 0x0000000;
		FFFInputState InputDown;
		InputDown.Buttons = 0x00000001;

		InputBuffer->AddInput(InputUp);
		InputBuffer->AddInput(InputDown);
		
		FFFInputCondition DownCondition;
		DownCondition.RequiredButtons = 0x00000001;
		DownCondition.RequiredButtonState = EFFButtonState::BTNS_Down;
		DownCondition.TimeoutDuration = 0;

		FFFInputSequence DownSequence;
		DownSequence.Sequence.Add(DownCondition);
		TestTrue(TEXT("Detect a single button press in the buffer"),
			InputBuffer->CheckInputSequence(DownSequence));
	}
	
	World->DestroyWorld(false);
	return true;
}

#endif //WITH_DEV_AUTOMATION_TESTS
